//-*-java-*-
/*
Copyright (c) 2003 Asoft ltd. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

  3. The names of the authors may not be used to endorse or promote products
     derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL ASOFT
LTD. OR ANY CONTRIBUTORS TO THIS SOFTWARE BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*                                                                             
*Copyright (c) 2003 Asoft ltd. All rights reserved.                            
*www.asoft.ru                                                                  
*Authors: Alexandre Rusev, Alexey Soloviev                                     
*/
package com.asoft.midp.dynarray;

import java.util.Vector;

public final class   char_array extends char_ArrayU implements char_AccessibleArray{
 private static final boolean FAST=true;
 private static final int DEBUG=1;    
 private int block_size=500;
 private int len;
 private final char[][] arrays;
 private final int max_blocks;
 private final int residue;
 private int total_allocated;
 private String name;    

 public char_array(int len){
  this(len,500,null);
 }
    
 public char_array(int len, int block_size){
  this(len,block_size,null);
 }

 public char_array(int len, String name){
  this(len,500,name); 
 }
    
 public char_array(int len, int block_size, String name){
  this.block_size=block_size;
  this.len=len;
  this.name=name;
  if(this.block_size<=0){this.block_size=100;}
  this.block_size=Math.min(this.block_size,len);
  if(len>=0){
   int n=this.len/this.block_size;
   residue=this.len-n*this.block_size;
   if(residue>0){++n;}
   max_blocks=n;
   arrays = new char[max_blocks][];
  }else{
   arrays = null;
   max_blocks=0;
   residue=0;
  }//if
 }

 public int length(){
  return len;     
 }
    
 public final int totalAllocated(){
  return total_allocated;
 }

 public int blockSize(){
  return block_size;
 }
    
 public final void set(int n, char val){
  final int c=n/block_size;
  if(arrays[c]==null&&val!=0){
   final int r=n%block_size;       
   final int m;  
   if(c==arrays.length-1&& this.residue>0){m=this.residue;}else{m=this.block_size;}
   arrays[c] = new char[m];
   total_allocated=total_allocated+m;
  }//if(arrays[c]==null)
  if(arrays[c]!=null){
   final int r=n%block_size;       
   arrays[c][r]=val;
  }//if
 }

 public final char get(int n){
  final int c=n/block_size;
  final int r=n%block_size;
  if(arrays[c]==null){
   return 0;   
  }else{
   return arrays[c][r];      
  }//if(arrays[c]==null)
 }

 /**  char[] src   char_array
  *
  */
  public final void copyFrom(char[] src,int srcoff, int dstoff, int len){
  if(FAST){
   //FAST==true
   Vector spl = new Vector();
   final int d=dstoff-srcoff;
   final int end=srcoff+len;
   //     " "
   // 
   int pd=(dstoff/this.block_size)*this.block_size - d;
   pd=pd+this.block_size;
   spl.addElement(new Integer(srcoff));
   while(pd<end){
    final int last=elem(spl.size()-1,spl);
    if(last<pd){spl.addElement(new Integer(pd));}
    pd=pd+this.block_size;
   }//wihile
   spl.addElement(new Integer(end));
   for(int i=0;i<spl.size()-1;++i){
    final int b=((Integer)spl.elementAt(i)).intValue();
    final int e=((Integer)spl.elementAt(i+1)).intValue()-1;
    final int l=e-b+1;
    final int dst_b=b+d;
    final int dst_c=(dst_b)/this.block_size;
    final int dst_r=(dst_b)%this.block_size;
    if(this.arrays[dst_c]==null){this.set(dst_b,(char)-1);}
    System.arraycopy(src,b,this.arrays[dst_c],dst_r,l);
   }//for
   
  }else{ 
   //FAST==false  
   for(int q=0;q<len;++q) {
    final int nsrc=srcoff+q;
    final int ndst=dstoff+q;
    this.set(ndst,src[nsrc]);
   }//for

  }//if(FAST)
   
 }

 /**   char_array  char[] dst
  *
  */
 public final void copyTo(int srcoff, char[] dst, int dstoff, int len){
   //FAST==true
  if(FAST){
   Vector spl = new Vector();     
   final int d=dstoff-srcoff;
   final int end=srcoff+len;
   int ps=(srcoff/this.block_size)*this.block_size;

   ps=ps+this.block_size;
   spl.addElement(new Integer(srcoff));
   while(ps<end){
    final int last=elem(spl.size()-1,spl);
    if(last<ps){spl.addElement(new Integer(ps));}
    ps=ps+this.block_size;
   }//wihile
   spl.addElement(new Integer(end));
   for(int i=0;i<spl.size()-1;++i){
    int b=((Integer)spl.elementAt(i)).intValue();
    int e=((Integer)spl.elementAt(i+1)).intValue()-1;
    int l=e-b+1;
    final int src_c=b/this.block_size;
    final int src_r=b%this.block_size;
    final int dst_b=b+d;
    if(this.arrays[src_c]!=null){
     System.arraycopy(this.arrays[src_c],src_r,dst,dst_b,l);	
    }else{
     clear(dst,dst_b,l);
    }//if
   }//for
   
     
  }else{ 
   //FAST=false
   for(int q=0;q<len;++q) {
    final int nsrc=srcoff+q;
    final int ndst=dstoff+q;      
    dst[ndst]=this.get(nsrc);
   }//for
   
  }//if(FAST)
 }

  public final void copyTo(int srcoff, char_AccessibleArray dst, int dstoff, int len){
   //FAST==true
  if(FAST){
   Vector spl = new Vector();     
   final int d=dstoff-srcoff;
   final int end=srcoff+len;
   int ps=(srcoff/this.block_size)*this.block_size;

   ps=ps+this.block_size;
   spl.addElement(new Integer(srcoff));
   while(ps<end){
    final int last=elem(spl.size()-1,spl);
    if(last<ps){spl.addElement(new Integer(ps));}
    ps=ps+this.block_size;
   }//wihile
   spl.addElement(new Integer(end));
   for(int i=0;i<spl.size()-1;++i){
    int b=((Integer)spl.elementAt(i)).intValue();
    int e=((Integer)spl.elementAt(i+1)).intValue()-1;
    int l=e-b+1;
    final int src_c=b/this.block_size;
    final int src_r=b%this.block_size;
    final int dst_b=b+d;
    if(this.arrays[src_c]!=null){
     //System.arraycopy(this.arrays[src_c],src_r,dst,dst_b,l);
     dst.copyFrom(this.arrays[src_c],src_r,dst_b,l);
    }else{
     dst.clear(dst_b,l);
    }//if
   }//for
   
     
  }else{ 
   //FAST=false
   for(int q=0;q<len;++q) {
    final int nsrc=srcoff+q;
    final int ndst=dstoff+q;      
    dst.set(  ndst,this.get(nsrc)  );
   }//for
   
  }//if(FAST)
 }

 public final void copyFrom(char_AccessibleArray src,int srcoff, int dstoff, int len){
  src.copyTo(srcoff,this,dstoff,len);
 }

 public final void clear(int off, int len){
  if(FAST) {
  //FAST==true
   char[] bb = new char[Math.min(50,this.length())];
   final int c=this.length()/bb.length;
   for(int i=0;i<=c;++i){
    final int off1=off+c*bb.length;
    if(off1<this.length()){
     //System.arraycopy(bb,0,a,off1,Math.min(bb.length,a.length-off1));	
     copyFrom(bb,0,off1,Math.min(bb.length,this.length()-off1));
    }//if(off1<this.length())
   }//for
  }else{
  //FAST==false
   for(int i=off;i<off+len;++i) {
    set(i,  (char)0  );
   }//for
  }//if(FAST)
 } 
//--------------------------------------------------------------------
 /*
 //    ,    char_array
 //       
 //     .
 //    char_array  char_AccessibleArray,
 //     .
 public static final void Copy(char_array src,int srcoff, char_array dst, int dstoff, int len){

  //if(len==0)return;     
  //src.copyTo(srcoff,dst,dstoff,len);
      
 if(len==0)return;

  if(FAST){
   //faster (but possibly more buggy) algorithm, using System.arraycopy
   //FAST==true
     
   Vector spl = new Vector();
     
   final int d=dstoff-srcoff;
   final int end=srcoff+len;
   //     " "
   // 
   int ps=(srcoff/src.block_size)*src.block_size;
   int pd=(dstoff/dst.block_size)*dst.block_size - d;

   ps=ps+src.block_size;
   pd=pd+dst.block_size;
   spl.addElement(new Integer(srcoff));
   while(pd<end || ps<end){
    final int last=elem(spl.size()-1,spl);
    if(ps<pd){
     if(last<ps){spl.addElement(new Integer(ps));}
     ps=ps+src.block_size;
    }else{
     if(last<pd){spl.addElement(new Integer(pd));}
     pd=pd+dst.block_size;	
    }//if
   }//wihile
   spl.addElement(new Integer(end));
   //       spl
   //     " "
   //   spl    
   //(    ! ;)  )
   //        ,
   //      end -   !
   

   

   //       "",
   //   ,    
   //    .
   //    (System.arraycopy)...
   //for(int i=0;i<spl.size();++i){
   // System.out.println(spl.elementAt(i));
   //}//for
    //System.out.println(); 
  final int A,B,C;
  if(srcoff<dstoff){
   // 
   A=spl.size()-2;B=0;C=-1;
  }else{
   // 
   A=0;B=spl.size()-2;C=1;
  }//if
  //System.out.println("!!!!!!!!!!!!!!! A="+A+" spl.size()="+spl.size()+" srcoff="+srcoff+" dstoff="+dstoff);
  for(int i=A;(C>0?i<=B:i>=B);i=i+C){
   //System.out.println("!!!!!!!!!!!!!!! i="+i+" A="+A+" B="+B);
   int b=((Integer)spl.elementAt(i)).intValue();
   int e=((Integer)spl.elementAt(i+1)).intValue()-1;
   int l=e-b+1;
   final int src_c=b/src.block_size;
   final int src_r=b%src.block_size;
   final int dst_b=b+d;
   final int dst_c=(dst_b)/dst.block_size;
   final int dst_r=(dst_b)%dst.block_size;

   //System.out.println("b="+b+" e="+e+" l="+l+":  src_c="+src_c+" src_r="+src_r+" dst_c="+dst_c+" dst_r="+dst_r);
   
  if(src.arrays[src_c]==null){
    if(dst.arrays[dst_c]!=null) {
     if(dst_r==0&&dst.arrays[dst_c].length==l){
      dst.arrays[dst_c]=null;
     }else{
      clear(dst.arrays[dst_c],dst_b,l);
     }//if
    }//if
   }else{
    //src.arrays[src_c]!=null   
    if(dst.arrays[dst_c]==null){dst.set(dst_b,(char)-1);}
     System.arraycopy(src.arrays[src_c],src_r,dst.arrays[dst_c],dst_r,l);      
   }//if
  
  //---DEBUG ONLY
   if(DEBUG>0){
    final int src_c_2=e/src.block_size;
    final int dst_c_2=(e+d)/dst.block_size;
    if(src_c!=src_c_2){System.out.println("dynamic_array: internal error #1");}
    if(dst_c!=dst_c_2){System.out.println("dynamic_array: internal error #1");}
   }//if(DEBUG>0)
   //---------
   }//for

  
  }else{
   //slower algorithm with no System.arrycopy using
   //FAST==false
   final int A,B,C;
   if(srcoff<dstoff){
    // 
    A=len-1;B=0;C=-1;
   }else{
    // 
    A=0;B=len-1;C=1;
   }//if

   for(int q=A;(C>0?q<=B:q>=B);q=q+C){   
   //for(int q=0;q<len;++q) {
    final int nsrc=srcoff+q;   
    final int csrc=nsrc/src.block_size;
    final int rsrc=nsrc%src.block_size;
       
    final int ndst=dstoff+q;   
    final int cdst=ndst/dst.block_size;
    final int rdst=ndst%dst.block_size;
 
    if(src.arrays[csrc]==null){
     if(dst.arrays[cdst]!=null){
      dst.set(ndst,(char)0);
     }//if
    }else{
     if(dst.arrays[cdst]!=null){
      dst.set(ndst,src.get(nsrc));
     }else
     if(src.get(nsrc)!=0){
      dst.set(ndst,src.get(nsrc));	
     }//if...else if ...
    }//if
   
    //dst.set(ndst,src.get(nsrc)); 
   }//for

  }//if(FAST)
  
 }

 public static final void Copy(char[] src,int srcoff, char_array dst, int dstoff, int len){
  if(FAST){
   //FAST==true
   Vector spl = new Vector();
   final int d=dstoff-srcoff;
   final int end=srcoff+len;
   //     " "
   // 
   int pd=(dstoff/dst.block_size)*dst.block_size - d;
   pd=pd+dst.block_size;
   spl.addElement(new Integer(srcoff));
   while(pd<end){
    final int last=elem(spl.size()-1,spl);
    if(last<pd){spl.addElement(new Integer(pd));}
    pd=pd+dst.block_size;
   }//wihile
   spl.addElement(new Integer(end));
   for(int i=0;i<spl.size()-1;++i){
    final int b=((Integer)spl.elementAt(i)).intValue();
    final int e=((Integer)spl.elementAt(i+1)).intValue()-1;
    final int l=e-b+1;
    final int dst_b=b+d;
    final int dst_c=(dst_b)/dst.block_size;
    final int dst_r=(dst_b)%dst.block_size;
    if(dst.arrays[dst_c]==null){dst.set(dst_b,(char)-1);}
    System.arraycopy(src,b,dst.arrays[dst_c],dst_r,l);
   }//for
   
  }else{ 
   //FAST==false  
   for(int q=0;q<len;++q) {
    final int nsrc=srcoff+q;
    final int ndst=dstoff+q;
    dst.set(ndst,src[nsrc]);
   }//for

  }//if(FAST)  
 }
  
 public static final void Copy(char_array src,int srcoff, char[] dst, int dstoff, int len){
   //FAST==true
  if(FAST){
   Vector spl = new Vector();     
   final int d=dstoff-srcoff;
   final int end=srcoff+len;
   int ps=(srcoff/src.block_size)*src.block_size;

   ps=ps+src.block_size;
   spl.addElement(new Integer(srcoff));
   while(ps<end){
    final int last=elem(spl.size()-1,spl);
    if(last<ps){spl.addElement(new Integer(ps));}
    ps=ps+src.block_size;
   }//wihile
   spl.addElement(new Integer(end));
   for(int i=0;i<spl.size()-1;++i){
    int b=((Integer)spl.elementAt(i)).intValue();
    int e=((Integer)spl.elementAt(i+1)).intValue()-1;
    int l=e-b+1;
    final int src_c=b/src.block_size;
    final int src_r=b%src.block_size;
    final int dst_b=b+d;
    if(src.arrays[src_c]!=null){
     System.arraycopy(src.arrays[src_c],src_r,dst,dst_b,l);	
    }else{
     clear(dst,dst_b,l);
    }//if
   }//for
   
     
  }else{ 
   //FAST=false
   for(int q=0;q<len;++q) {
    final int nsrc=srcoff+q;
    final int ndst=dstoff+q;      
    dst[ndst]=src.get(nsrc);
   }//for
   
  }//if(FAST)
 } 
 */ 
//--------------------------------------------------------
private static final void clear(char[] a,int off, int len){
 //System.out.println("fill in array with "+len+" zeros");
 if(FAST) {
 //FAST==true
  char[] bb = new char[Math.min(50,a.length)];
  final int c=a.length/bb.length;
  for(int i=0;i<=c;++i){
   final int off1=off+c*bb.length;
   if(off1<a.length)System.arraycopy(bb,0,a,off1,Math.min(bb.length,a.length-off1));   
  }//for
 }else{
 //FAST==false
  for(int i=off;i<off+len;++i) {
   a[i]=(char)0;
  }//for
 }//if(FAST)
}
//--------------------------------------------------------
 private static final int elem(int n,Vector vec){
  return ((Integer)vec.elementAt(n)).intValue();   
 }
//--------------------------------------------------------
 // protected void finalize(){
 //  System.out.println("Dynamic array of char "+name+" used "+totalAllocated()+" of "+length()+" allocated  ("+(totalAllocated()*100/length())+"%)");
 // }
//========================================================

 public static void main(String[] args){


 char_array src = new char_array(100,3);
 char_array dst = new char_array(100,4);
 int srcoff=3;
 int dstoff=3;
 int len=9;
 char_array.Copy(src,srcoff, dst,dstoff,len);
/*     
  final int n=65536;
  char_array array = new char_array(n);
  for(int i=0;i<n;++i){
   char q=(char)i;   
   array.set(i,q);
   if(q!=array.get(i)){
    System.out.print(i+": ");       
    System.out.println(" error "+q+"!="+array.get(i));
    System.out.println();    
   }
  }//for

  for(int i=0;i<n;++i){
   char q=(char)i;
   if(q!=array.get(i)){
    System.out.print(i+": ");       
    System.out.println(" error "+q+"!="+array.get(i));
    System.out.println();    
   }
  }//for
*/  
 }

    
    

}

